import sys
import json
import yaml
from typing import Dict, List, Tuple, Union

import typer
from rich import print_json
from rich.table import Table

from monarch_py.datamodels.model import (
    AssociationCountList,
    ConfiguredBaseModel,
    Entity,
    HistoPheno,
    Node,
    Release,
    Results,
)
from monarch_py.utils.utils import console

FMT_INPUT_ERROR_MSG = """Text conversion method only accepts the following object types:
    Entity
    HistoPheno
    AssociationCountList
    Release
    Results
"""
# VALID_MODELS = [Entity, HistoPheno, AssociationCountList, Release, Results]


### Converters ###


def get_headers_from_obj(obj: ConfiguredBaseModel) -> list:
    """Return a list of headers from a pydantic model."""
    if isinstance(obj, Entity):
        headers = obj.model_dump().keys()
    elif isinstance(obj, (AssociationCountList, HistoPheno, Results)):
        if obj.items:
            headers = obj.items[0].model_dump().keys()
        else:
            schema = type(obj).model_json_schema()
            definitions = schema["definitions"]
            this_ref = schema["properties"]["items"]["items"]["$ref"].split("/")[-1]
            headers = definitions[this_ref]["properties"].keys()
    else:
        console.print(f"\n[bold red]{FMT_INPUT_ERROR_MSG}[/]")
        raise typer.Exit(1)
    return list(headers)


def get_headers_and_rows(obj: ConfiguredBaseModel) -> Tuple[str, str]:
    """Converts a pydantic model to a TSV string."""
    if isinstance(obj, (Entity, Release)):
        headers = obj.model_dump().keys()
        rows = [list(obj.model_dump().values())]
    elif isinstance(obj, (AssociationCountList, HistoPheno, Results)):
        if not obj.items:
            headers = get_headers_from_obj(obj)
            rows = []
        else:
            headers = obj.items[0].model_dump().keys()
            rows = [list(item.model_dump().values()) for item in obj.items]
    else:
        console.print(f"\n[bold red]{FMT_INPUT_ERROR_MSG}[/]")
        raise typer.Exit(1)
    return headers, rows


### Printers/writers ###


def to_json(
    obj: Union[ConfiguredBaseModel, Dict, List[ConfiguredBaseModel]], file: str = None, print_output: bool = True
) -> str:
    """Converts a pydantic model to a JSON string."""
    if isinstance(obj, ConfiguredBaseModel):
        json_value = obj.model_dump_json(indent=4)
    elif isinstance(obj, dict):
        json_value = json.dumps(obj, indent=4)
    elif isinstance(obj, list):
        # json_value = json.dumps({"items": [o.model_dump_json() for o in obj]}, indent=4)
        json_value = json.dumps([json.loads(o.model_dump_json(indent=4)) for o in obj])
    if file:
        with open(file, "w") as f:
            f.write(json_value)
        console.print(f"\nOutput written to {file}\n")
    elif print_output and not file:
        print_json(json_value)
    return json_value


def to_tsv(obj: ConfiguredBaseModel, file: str = None, print_output: bool = True) -> str:
    """Prints or writes a pydantic model to a TSV file/stdout."""
    headers, rows = get_headers_and_rows(obj)
    tsv = ""
    tsv += "\t".join(headers)
    for row in rows:
        row = "\t".join([str(r) for r in row])
        tsv += f"\n{row}"
    if file:
        with open(file, "w") as f:
            f.write(tsv)
        console.print(f"\nOutput written to {file}\n")
    elif print_output:
        console.print(tsv)
    return tsv


def to_table(obj: ConfiguredBaseModel, print_output: bool = True) -> Table:
    """Prints a pydantic model as a rich Table."""
    if isinstance(obj, Node):
        console.print(f"\n[bold red]Table output not implemented for Node objects.[/]\n")
        raise typer.Exit(1)
    headers, rows = get_headers_and_rows(obj)
    for row in rows:
        for i, value in enumerate(row):
            if isinstance(value, list) and all(isinstance(v, str) for v in value):
                row[i] = ", ".join(value)
            elif not isinstance(value, str):
                row[i] = str(value)
            else:
                row[i] = value
    title = f"{obj.__class__.__name__}: {obj.id}" if hasattr(obj, "id") else obj.__class__.__name__
    table = Table(
        title=console.rule(title),
        show_header=True,
        header_style="bold cyan",
    )
    for header in headers:
        table.add_column(header)
    for row in rows:
        table.add_row(*row)
    if print_output:
        console.print(table)
    return table


def to_yaml(obj: ConfiguredBaseModel, file: str = None, print_output: bool = True) -> Union[Dict, List[Dict]]:
    """Converts a pydantic model to a YAML string."""
    if isinstance(obj, (Entity, Release)):
        o = obj.model_dump()
    elif isinstance(obj, (Results, HistoPheno, AssociationCountList)):
        o = [item.model_dump() for item in obj.items]
    else:
        console.print(f"\n[bold red]{FMT_INPUT_ERROR_MSG}[/]\n")
        raise typer.Exit(1)
    if file:
        fh = open(file, "w")
        yaml.dump(o, fh, indent=4)
        console.print(f"\nOutput written to {file}\n")
        fh.close()
    elif print_output and not file:
        yaml.dump(o, sys.stdout, indent=4)
    return yaml.dump(o)


def format_output(fmt: str, response: Union[ConfiguredBaseModel, Dict], output: str = None):
    if fmt.lower() == "json":
        to_json(response, output)
    elif fmt.lower() == "tsv":
        to_tsv(response, output)
    elif fmt.lower() == "yaml":
        to_yaml(response, output)
    elif fmt.lower() == "table":
        to_table(response)
    else:
        console.print(f"\n[bold red]Format '{fmt}' not supported.[/]\n")
        raise typer.Exit(1)



import os
from dataclasses import dataclass
from typing import List, Union, Optional

import requests
from monarch_py.datamodels.model import (
    Association,
    CompactAssociation,
    AssociationCountList,
    AssociationResults,
    AssociationTableResults,
    CategoryGroupedAssociationResults,
    Entity,
    HistoPheno,
    MappingResults,
    MultiEntityAssociationResults,
    Node,
    NodeHierarchy,
    SearchResults,
)
from monarch_py.datamodels.solr import core
from monarch_py.datamodels.category_enums import (
    AssociationCategory,
    AssociationPredicate,
    EntityCategory,
    MappingPredicate,
)
from monarch_py.implementations.solr.solr_parsers import (
    convert_facet_fields,
    convert_facet_queries,
    parse_association_counts,
    parse_association_table,
    parse_associations,
    parse_autocomplete,
    parse_entity,
    parse_histopheno,
    parse_mappings,
    parse_search,
)
from monarch_py.implementations.solr.solr_query_utils import (
    build_association_counts_query,
    build_association_query,
    build_association_table_query,
    build_autocomplete_query,
    build_histopheno_query,
    build_mapping_query,
    build_multi_entity_association_query,
    build_search_query,
    build_grounding_query,
)
from monarch_py.interfaces.association_interface import AssociationInterface
from monarch_py.interfaces.entity_interface import EntityInterface
from monarch_py.interfaces.search_interface import SearchInterface
from monarch_py.interfaces.grounding_interface import GroundingInterface
from monarch_py.service.solr_service import SolrService
from monarch_py.utils.entity_utils import get_expanded_curie, get_uri
from monarch_py.utils.utils import get_provided_by_link, get_links_for_field


@dataclass
class SolrImplementation(EntityInterface, AssociationInterface, SearchInterface, GroundingInterface):
    """Implementation of Monarch Interfaces for Solr endpoint"""

    base_url: str = os.getenv("MONARCH_SOLR_URL", "http://localhost:8983/solr")

    def solr_is_available(self) -> bool:
        """Check if the Solr instance is available"""
        try:
            return requests.get(self.base_url).status_code == 200
        except Exception:
            return False

    ###############################
    # Implements: EntityInterface #
    ###############################

    def get_entity(self, id: str, extra: bool) -> Optional[Union[Node, Entity]]:
        """Retrieve a specific entity by exact ID match, with optional extras

        Args:
            id (str): id of the entity to search for.
            extra (bool): Whether to include association counts and hierarchy in the response.

        Returns:
            Entity: Dataclass representing results of an entity search.
            Node: Dataclass representing results of an entity search with extra=True.
        """
        solr = SolrService(base_url=self.base_url, core=core.ENTITY)
        solr_document = solr.get(id)
        if solr_document is None:
            return None
        if not extra:
            return parse_entity(solr_document)

        # Get extra data (this logic is very tricky to test because of the calls to Solr)
        entity = Entity(**solr_document)
        entity.uri = get_uri(entity.id)
        if "biolink:Disease" == entity.category:
            # Get mode of inheritance
            mode_of_inheritance_associations = self.get_associations(
                subject=id, predicate=[AssociationPredicate.HAS_MODE_OF_INHERITANCE], direct=True, offset=0
            )
            if mode_of_inheritance_associations is not None and len(mode_of_inheritance_associations.items) == 1:
                entity.inheritance = self._get_counterpart_entity(mode_of_inheritance_associations.items[0], entity)
            # Get causal gene
            entity.causal_gene = [
                self._get_counterpart_entity(association, entity)
                for association in self.get_associations(
                    object=id,
                    direct=True,
                    predicate=[AssociationPredicate.CAUSES],
                    category=[AssociationCategory.CAUSAL_GENE_TO_DISEASE_ASSOCIATION],
                ).items
            ]
        if "biolink:Gene" == entity.category:
            entity.causes_disease = [
                self._get_counterpart_entity(association, entity)
                for association in self.get_associations(
                    subject=id,
                    direct=True,
                    predicate=[AssociationPredicate.CAUSES],
                    category=[AssociationCategory.CAUSAL_GENE_TO_DISEASE_ASSOCIATION],
                ).items
            ]
        node: Node = Node(
            **entity.model_dump(),
            node_hierarchy=self._get_node_hierarchy(entity),
            association_counts=self.get_association_counts(id).items,
            external_links=get_links_for_field(entity.xref) if entity.xref else [],
            provided_by_link=get_provided_by_link(entity.provided_by),
            mappings=self._get_mapped_entities(entity),
        )

        return node

    ### Entity helpers ###

    def _get_mapped_entities(self, this_entity: Entity) -> list:
        """..."""
        mapped_entities = []
        mappings = self.get_mappings(entity_id=this_entity.id)
        for m in mappings.items:
            if this_entity.id == m.subject_id:
                mapped_entities.append(get_expanded_curie(m.object_id))
            elif this_entity.id == m.object_id:
                mapped_entities.append(get_expanded_curie(m.subject_id))
            else:
                pass
        return mapped_entities

    def _get_counterpart_entity(self, association: Association, this_entity: Entity) -> Entity:
        """Returns the id, name, and category of the other Entity in an Association given this_entity"""
        if this_entity.id == association.subject:
            entity = Entity(
                id=association.object,
                name=association.object_label,
                category=association.object_category,
            )
        elif this_entity.id == association.object:
            entity = Entity(
                id=association.subject,
                name=association.subject_label,
                category=association.subject_category,
            )
        else:
            raise ValueError(f"Association does not contain this_entity: {this_entity.id}")

        return entity

    def get_counterpart_entities(
        self,
        this_entity: Entity,
        entity: Optional[str] = None,
        subject: Optional[str] = None,
        predicate: List[AssociationPredicate] = None,
        object: Optional[str] = None,
    ) -> List[Entity]:
        """
        Get a list of entities directly associated with this_entity fetched from associations
        in the Solr index

        Args:
            this_entity (Entity): The entity to get associations for
            entity (str, optional): an entity ID occurring in either the subject or object. Defaults to None.
            subject (str, optional): an entity ID occurring in the subject. Defaults to None.
            predicate (str, optional): a predicate value. Defaults to None.
            object (str, optional): an entity ID occurring in the object. Defaults to None.
        """
        return [
            self._get_counterpart_entity(association, this_entity)
            for association in self.get_associations(
                entity=entity,
                subject=subject,
                predicate=predicate,
                object=object,
                direct=True,
                limit=1000,
                offset=0,
            ).items
        ]

    def _get_node_hierarchy(self, entity: Entity) -> NodeHierarchy:
        """
        Get a NodeHierarchy for the given entity

        Args:
            entity (Entity): The entity to get the hierarchy for
            si (SolrInterface): A SolrInterface instance

        Returns:
            NodeHierarchy: A NodeHierarchy object
        """

        super_classes = self.get_counterpart_entities(
            this_entity=entity, subject=entity.id, predicate=[AssociationPredicate.SUBCLASS_OF]
        )
        sub_classes = self.get_counterpart_entities(
            this_entity=entity, object=entity.id, predicate=[AssociationPredicate.SUBCLASS_OF]
        )

        return NodeHierarchy(
            super_classes=super_classes,
            sub_classes=sub_classes,
        )

    ####################################
    # Implements: AssociationInterface #
    ####################################

    def get_associations(
        self,
        category: List[AssociationCategory] = None,
        subject: Optional[List[str]] = None,
        subject_closure: Optional[str] = None,
        subject_category: List[EntityCategory] = None,
        subject_namespace: Optional[List[str]] = None,
        subject_taxon: Optional[List[str]] = None,
        predicate: List[AssociationPredicate] = None,
        object: Optional[List[str]] = None,
        object_closure: Optional[str] = None,
        object_category: List[EntityCategory] = None,
        object_namespace: Optional[List[str]] = None,
        object_taxon: Optional[List[str]] = None,
        entity: Optional[List[str]] = None,
        direct: bool = False,
        q: Optional[str] = None,
        facet_fields: Optional[List[str]] = None,
        facet_queries: Optional[List[str]] = None,
        filter_queries: Optional[List[str]] = None,
        compact: bool = False,
        offset: int = 0,
        limit: int = 20,
    ) -> Union[AssociationResults, CompactAssociation]:
        """Retrieve paginated association records, with filter options

        Args:
            category: Filter to only associations matching the specified categories. Defaults to None.
            predicate: Filter to only associations matching the specified predicates. Defaults to None.
            subject: Filter to only associations matching the specified subjects. Defaults to None.
            object: Filter to only associations matching the specified objects. Defaults to None.
            subject_closure: Filter to only associations with the specified term ID as an ancestor of the subject. Defaults to None
            object_closure: Filter to only associations with the specified term ID as an ancestor of the object. Defaults to None
            entity: Filter to only associations where the specified entities are the subject or the object. Defaults to None.
            q: Query string to search within matches. Defaults to None.
            facet_fields: List of fields to include facet counts for. Defaults to None.
            facet_queries: List of queries to include facet counts for. Defaults to None.
            filter_queries: List of queries to filter results by. Defaults to None.
            compact: Return compact results with fewer fields. Defaults to False.
            offset: Result offset, for pagination. Defaults to 0.
            limit: Limit results to specified number. Defaults to 20.

        Returns:
            AssociationResults: Dataclass representing results of an association search.
        """
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)
        query = build_association_query(
            category=[c.value for c in category] if category else [],
            predicate=[p.value for p in predicate] if predicate else [],
            subject=[subject] if isinstance(subject, str) else subject,
            object=[object] if isinstance(object, str) else object,
            entity=[entity] if isinstance(entity, str) else entity,
            subject_closure=subject_closure,
            object_closure=object_closure,
            subject_category=[c.value for c in subject_category] if subject_category else [],
            subject_namespace=[subject_namespace] if isinstance(subject_namespace, str) else subject_namespace,
            subject_taxon=[subject_taxon] if isinstance(subject_taxon, str) else subject_taxon,
            object_category=[c.value for c in object_category] if object_category else [],
            object_taxon=[object_taxon] if isinstance(object_taxon, str) else object_taxon,
            object_namespace=[object_namespace] if isinstance(object_namespace, str) else object_namespace,
            direct=direct,
            q=q,
            facet_fields=facet_fields,
            facet_queries=facet_queries,
            filter_queries=filter_queries,
            offset=offset,
            limit=limit,
        )
        query_result = solr.query(query)

        associations = parse_associations(query_result, compact, offset, limit)
        return associations

    def get_histopheno(self, subject: Optional[str] = None) -> HistoPheno:
        """Get histopheno counts for a given subject_closure"""
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)
        query = build_histopheno_query(subject)
        query_result = solr.query(query)
        histopheno = parse_histopheno(query_result, subject)
        return histopheno

    def get_multi_entity_associations(
        self,
        entity: List[str],
        counterpart_category: Optional[List[str]] = None,
        offset: int = 0,
        # limit: int = 20,
        limit_per_group: int = 20,
    ) -> List[MultiEntityAssociationResults]:
        """Get associations between multiple entities and counterparts of a given category

        Args:
            entity (List[str]): List of entity IDs to get associations for
            counterpart_category (List[str], optional): List of categories of counterpart entity to get associations for. Defaults to None.
            offset (int, optional): Result offset, for pagination. Defaults to 0.
            limit (int, optional): Limit results to specified number. Defaults to 20.
            limit_per_group (int, optional): Limit results to specified number per group. Defaults to 20.
        """
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)
        results = []
        for entity_id in entity:
            ent = self.get_entity(entity_id, extra=False)
            if ent is None:
                results.append(
                    MultiEntityAssociationResults(
                        id=entity_id, name="Entity not found", total=0, offset=0, limit=0, associated_categories=[]
                    )
                )
                continue
            entity_result = MultiEntityAssociationResults(
                id=ent.id, name=ent.name, total=0, offset=offset, limit=limit_per_group, associated_categories=[]
            )
            if counterpart_category:
                for category in counterpart_category:
                    query = build_multi_entity_association_query(
                        entity=ent.id, counterpart_category=category, offset=offset, limit=limit_per_group
                    )
                    query_result = solr.query(query)
                    associations = parse_associations(query_result)
                    entity_result.associated_categories.append(
                        CategoryGroupedAssociationResults(
                            counterpart_category=category,
                            items=associations.items,
                            total=associations.total,
                            offset=associations.offset,
                            limit=associations.limit,
                        )
                    )
                    entity_result.total += associations.total
            results.append(entity_result)
        return results

    ###############################
    # Implements: SearchInterface #
    ###############################

    def search(
        self,
        q: str = "*:*",
        category: Union[List[EntityCategory], None] = None,
        in_taxon_label: Union[List[str], None] = None,
        facet_fields: Union[List[str], None] = None,
        facet_queries: Union[List[str], None] = None,
        filter_queries: Union[List[str], None] = None,
        sort: Optional[str] = None,
        highlighting: bool = False,
        offset: int = 0,
        limit: int = 20,
    ) -> SearchResults:
        """Search for entities by label, with optional filters

        Args:
            q (str): Query string. Defaults to "*:*".
            offset (int): Result offset, for pagination. Defaults to 0.
            limit (int): Limit results to specified number. Defaults to 20.
            category (List[str]): Filter to only entities matching the specified categories. Defaults to None.
            in_taxon_label (List[str]): Filter to only entities matching the specified taxon label. Defaults to None.
            facet_fields (List[str]): List of fields to include facet counts for. Defaults to None.
            facet_queries (List[str]): List of queries to include facet counts for. Defaults to None.
            filter_queries (List[str]): List of queries to filter results by. Defaults to None.
            sort (str): Sort results by the specified field. Defaults to None.

        Returns:
            SearchResults: Dataclass representing results of a search.
        """
        query = build_search_query(
            q=q,
            category=[c.value for c in category] if category else None,
            in_taxon_label=in_taxon_label,
            facet_fields=facet_fields,
            facet_queries=facet_queries,
            filter_queries=filter_queries,
            highlighting=highlighting,
            sort=sort,
            offset=offset,
            limit=limit,
        )
        solr = SolrService(base_url=self.base_url, core=core.ENTITY)
        query_result = solr.query(query)
        results = parse_search(query_result)
        return results

    def autocomplete(
        self, q: str, category: List[EntityCategory] = None, prioritized_predicates: List[AssociationPredicate] = None
    ) -> SearchResults:
        solr = SolrService(base_url=self.base_url, core=core.ENTITY)
        query = build_autocomplete_query(
            q,
            category=[cat.value for cat in category] if category else None,
            prioritized_predicates=prioritized_predicates,
        )
        query_result = solr.query(query)
        results = parse_autocomplete(query_result)
        return results

    def get_association_counts(self, entity: str) -> AssociationCountList:
        """Get list of association counts for a given entity"""
        query = build_association_counts_query(entity)
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)
        query_result = solr.query(query)
        association_counts = parse_association_counts(query_result, entity)
        return association_counts

    def get_association_facets(
        self,
        category: List[AssociationCategory] = None,
        subject: Optional[List[str]] = None,
        predicate: List[AssociationPredicate] = None,
        object: Optional[List[str]] = None,
        subject_closure: Optional[str] = None,
        object_closure: Optional[str] = None,
        entity: Optional[List[str]] = None,
        facet_fields: Optional[List[str]] = None,
        facet_queries: Optional[List[str]] = None,
        filter_queries: Optional[List[str]] = None,
    ) -> SearchResults:
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)

        query = build_association_query(
            category=[c.value for c in category] if category else None,
            predicate=[p.value for p in predicate] if predicate else None,
            subject=[subject] if isinstance(subject, str) else subject,
            object=[object] if isinstance(object, str) else object,
            entity=[entity] if isinstance(entity, str) else entity,
            subject_closure=subject_closure,
            object_closure=object_closure,
            offset=0,
            limit=0,
            facet_fields=facet_fields,
            facet_queries=facet_queries,
            filter_queries=filter_queries,
        )
        query_result = solr.query(query)
        return SearchResults(
            limit=0,
            offset=0,
            total=query_result.response.num_found,
            items=[],
            facet_fields=(
                convert_facet_fields(query_result.facet_counts.facet_fields) if query_result.facet_counts else []
            ),
            facet_queries=(
                convert_facet_queries(query_result.facet_counts.facet_queries) if query_result.facet_counts else []
            ),
        )

    def get_association_table(
        self,
        entity: str,
        category: AssociationCategory,
        traverse_orthologs: bool = False,
        direct: bool = False,
        q: Optional[str] = None,
        facet_fields: Optional[List[str]] = None,
        facet_queries: Optional[List[str]] = None,
        filter_queries: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
        offset: int = 0,
        limit: int = 5,
    ) -> AssociationTableResults:
        entities = [entity]
        if traverse_orthologs:
            ortholog_associations = self.get_associations(
                entity=[entity], predicate=[AssociationPredicate.ORTHOLOGOUS_TO]
            )
            orthologous_entities = [
                self._get_counterpart_entity(a, Entity(id=entity)) for a in ortholog_associations.items
            ]
            entities.extend([ent.id for ent in orthologous_entities])
        query = build_association_table_query(
            entity=entities,
            category=category.value,
            direct=direct,
            q=q,
            facet_fields=facet_fields,
            facet_queries=facet_queries,
            filter_queries=filter_queries,
            sort=sort,
            offset=offset,
            limit=limit,
        )
        solr = SolrService(base_url=self.base_url, core=core.ASSOCIATION)
        query_result = solr.query(query)
        return parse_association_table(query_result, entities, offset, limit)

    def get_mappings(
        self,
        entity_id: Optional[List[str]] = None,
        subject_id: Optional[List[str]] = None,
        predicate_id: Optional[List[MappingPredicate]] = None,
        object_id: Optional[List[str]] = None,
        mapping_justification: Optional[List[str]] = None,
        offset: int = 0,
        limit: int = 20,
    ) -> MappingResults:
        solr = SolrService(base_url=self.base_url, core=core.SSSOM)
        query = build_mapping_query(
            entity_id=[entity_id] if isinstance(entity_id, str) else entity_id,
            subject_id=[subject_id] if isinstance(subject_id, str) else subject_id,
            predicate_id=[p.value for p in predicate_id] if predicate_id else None,
            object_id=[object_id] if isinstance(object_id, str) else object_id,
            mapping_justification=(
                [mapping_justification] if isinstance(mapping_justification, str) else mapping_justification
            ),
            offset=offset,
            limit=limit,
        )
        query_result = solr.query(query)
        mappings = parse_mappings(query_result, offset, limit)
        return mappings

    ##################################
    # Implements: GroundingInterface #
    ##################################

    def ground_entity(self, text: str) -> List[Entity]:
        """Grounds a single entity

        Args:
            text (str): Text to ground

        Returns:
            Entity: Dataclass representing a single entity
        """
        solr = SolrService(base_url=self.base_url, core=core.ENTITY)
        query = build_grounding_query(text)
        query_result = solr.query(query)
        search_result = parse_search(query_result)
        entities = [entity for entity in search_result.items[:3]]
        return entities



from typing import Dict, List

from loguru import logger
from pydantic import ValidationError

from monarch_py.datamodels.model import (
    Association,
    CompactAssociation,
    CompactAssociationResults,
    AssociationCount,
    AssociationCountList,
    AssociationDirectionEnum,
    AssociationResults,
    AssociationTableResults,
    DirectionalAssociation,
    Entity,
    FacetField,
    FacetValue,
    HistoBin,
    HistoPheno,
    Mapping,
    MappingResults,
    SearchResult,
    SearchResults,
)
from monarch_py.datamodels.solr import HistoPhenoKeys, SolrQueryResult
from monarch_py.service.curie_service import converter
from monarch_py.utils.association_type_utils import get_association_type_mapping_by_query_string
from monarch_py.utils.utils import get_links_for_field, get_provided_by_link

####################
# Parser functions #
####################


def parse_associations(
    query_result: SolrQueryResult,
    compact: bool = False,
    offset: int = 0,
    limit: int = 20,
) -> AssociationResults:
    associations = []
    total = query_result.response.num_found
    if compact:
        associations = [
            CompactAssociation(
                category=doc.get("category"),
                subject=doc.get("subject"),
                subject_label=doc.get("subject_label"),
                predicate=doc.get("predicate"),
                object=doc.get("object"),
                object_label=doc.get("object_label"),
                negated=doc.get("negated"),
            )
            for doc in query_result.response.docs
        ]
        return CompactAssociationResults(
            items=associations,
            limit=limit,
            offset=offset,
            total=total,
            facet_fields=convert_facet_fields(query_result.facet_counts.facet_fields),
            facet_queries=convert_facet_queries(query_result.facet_counts.facet_queries),
        )
    else:
        for doc in query_result.response.docs:
            try:
                association = Association(**doc)
            except ValidationError:
                logger.error(f"Validation error for {doc}")
                raise ValidationError
            association.provided_by_link = (
                get_provided_by_link(association.provided_by) if association.provided_by else []
            )
            association.has_evidence_links = (
                get_links_for_field(association.has_evidence) if association.has_evidence else []
            )
            association.publications_links = (
                get_links_for_field(association.publications) if association.publications else []
            )
            associations.append(association)
        return AssociationResults(
            items=associations,
            limit=limit,
            offset=offset,
            total=total,
            facet_fields=convert_facet_fields(query_result.facet_counts.facet_fields),
            facet_queries=convert_facet_queries(query_result.facet_counts.facet_queries),
        )


def parse_association_counts(query_result: SolrQueryResult, entity: str) -> AssociationCountList:
    subject_query = f'AND (subject:"{entity}" OR subject_closure:"{entity}")'
    object_query = f'AND (object:"{entity}" OR object_closure:"{entity}" OR disease_context_qualifier:"{entity}" OR disease_context_qualifier_closure:"{entity}")'
    association_count_dict: Dict[str, AssociationCount] = {}
    for k, v in query_result.facet_counts.facet_queries.items():
        if v > 0:
            if k.endswith(subject_query):
                original_query = k.replace(f" {subject_query}", "").lstrip("(").rstrip(")")
                agm = get_association_type_mapping_by_query_string(original_query)
                label = agm.subject_label
            elif k.endswith(object_query):
                original_query = k.replace(f" {object_query}", "").lstrip("(").rstrip(")")
                agm = get_association_type_mapping_by_query_string(original_query)
                label = agm.object_label
                # always use forward for symmetric association types
            else:
                raise ValueError(f"Unexpected facet query when building association counts: {k}")
            # Symmetric associations need to be summed together, since both directions will be returned
            # when searching for associations by type
            if label in association_count_dict and agm.symmetric:
                association_count_dict[label].count += v
            else:
                association_count_dict[label] = AssociationCount(
                    label=label,
                    count=v,
                    category=agm.category,
                )
    return AssociationCountList(items=list(association_count_dict.values()))


def parse_entity(solr_document: Dict) -> Entity:
    try:
        entity = Entity(**solr_document)
        entity.uri = converter.expand(entity.id)
    except ValidationError:
        logger.error(f"Validation error for {solr_document}")
        raise
    return entity


def parse_association_table(
    query_result: SolrQueryResult,
    entity: List[str],
    offset: int,
    limit: int,
) -> AssociationTableResults:
    total = query_result.response.num_found
    associations: List[DirectionalAssociation] = []
    for doc in query_result.response.docs:
        try:
            direction = get_association_direction(entity, doc)
            association = DirectionalAssociation(**doc, direction=direction)
            association.provided_by_link = (
                get_provided_by_link(association.provided_by) if association.provided_by else []
            )
            association.has_evidence_links = (
                get_links_for_field(association.has_evidence) if association.has_evidence else []
            )
            association.publications_links = (
                get_links_for_field(association.publications) if association.publications else []
            )
            associations.append(association)
        except ValidationError:
            logger.error(f"Validation error for {doc}")
            raise
    results = AssociationTableResults(
        items=associations,
        limit=limit,
        offset=offset,
        total=total,
        facet_fields=convert_facet_fields(query_result.facet_counts.facet_fields),
        facet_queries=convert_facet_queries(query_result.facet_counts.facet_queries),
    )
    for i in zip(results.items, query_result.response.docs):
        assert i[0].subject == i[1]["subject"]
    return results


def parse_histopheno(query_result: SolrQueryResult, subject_closure: str) -> HistoPheno:
    """Parse a SolrQueryResult into a HistoPheno object"""
    bins = []
    for k, v in query_result.facet_counts.facet_queries.items():
        id = f"{k.split(':')[1]}:{k.split(':')[2]}".replace('"', "")
        label = HistoPhenoKeys(id).name
        bins.append(HistoBin(id=id, label=label, count=v))
    bins = sorted(bins, key=lambda x: x.count, reverse=True)

    return HistoPheno(id=subject_closure, items=bins)


def parse_search(
    query_result: SolrQueryResult,
    offset: int = 0,
    limit: int = 20,
) -> SearchResults:
    items = []
    for doc in query_result.response.docs:
        try:
            result = SearchResult(**doc)
            items.append(result)
        except ValidationError:
            logger.error(f"Validation error for {doc}")
            raise
    total = query_result.response.num_found
    facet_fields = convert_facet_fields(query_result.facet_counts.facet_fields)
    facet_queries = convert_facet_queries(query_result.facet_counts.facet_queries)
    return SearchResults(
        limit=limit, offset=offset, total=total, items=items, facet_fields=facet_fields, facet_queries=facet_queries
    )


def parse_autocomplete(query_result: SolrQueryResult) -> SearchResults:
    total = query_result.response.num_found
    items = []
    for doc in query_result.response.docs:
        try:
            result = SearchResult(**doc)
            items.append(result)
        except ValidationError:
            logger.error(f"Validation error for {doc}")
            raise
    return SearchResults(limit=10, offset=0, total=total, items=items)


def parse_mappings(query_result: SolrQueryResult, offset: int = 0, limit: int = 20) -> MappingResults:
    total = query_result.response.num_found
    items = []
    for doc in query_result.response.docs:
        try:
            result = Mapping(**doc)
            items.append(result)
        except ValidationError:
            logger.error(f"Validation error for {doc}")
            raise
    return MappingResults(limit=limit, offset=offset, total=total, items=items)


##################
# Parser Helpers #
##################


def convert_facet_fields(solr_facet_fields: Dict) -> List[FacetField]:
    """Converts list of raw Solr facet fields FacetField instances"""
    facet_fields: List[FacetField] = []
    for field in solr_facet_fields:
        ff = FacetField(label=field)
        facet_list = solr_facet_fields[field]
        facet_dict = dict(zip(facet_list[::2], facet_list[1::2]))
        ff.facet_values = [FacetValue(label=k, count=v) for k, v in facet_dict.items()]
        facet_fields.append(ff)
    return facet_fields


def convert_facet_queries(solr_facet_queries: Dict[str, int]) -> List[FacetValue]:
    """Converts list of raw Solr facet queries to FacetValue instances"""
    return [FacetValue(label=k, count=v) for k, v in solr_facet_queries.items()]


def get_association_direction(entity: List[str], document: Dict) -> AssociationDirectionEnum:
    """Get the direction of an association based on the entity and the association document"""
    if document.get("subject") in entity or (
        document.get("subject_closure") and any(e in document.get("subject_closure") for e in entity)
    ):
        direction = AssociationDirectionEnum.outgoing
    elif document.get("object") in entity or (
        document.get("object_closure") and any(e in document.get("object_closure") for e in entity)
    ):
        direction = AssociationDirectionEnum.incoming
    elif document.get("disease_context_qualifier") in entity or (
        document.get("disease_context_qualifier_closure")
        and any(e in document.get("disease_context_qualifier_closure") for e in entity)
    ):
        # This is a special case for disease_context_qualifier, if an association between two other entities
        # only occurs within the context of a disease, we can treat it like an incoming association
        direction = AssociationDirectionEnum.incoming
    else:
        raise ValueError(f"Entity {entity} not found in association {document}")
    return direction



import os
import shutil
import sys
import time

import docker
import pystow
import typer
from monarch_py.implementations.solr.solr_implementation import SolrImplementation
from monarch_py.utils.utils import KG_DEV_URL, console

monarchstow = pystow.module("monarch")


def check_for_docker():
    try:
        docker.from_env()
    except docker.errors.DockerException:
        typer.secho(
            f"\n\tDocker not found\n\tPlease install Docker, and refer to:",
            fg=typer.colors.RED,
        )
        typer.secho(
            f"\tmonarch solr --help\n",
            fg=typer.colors.WHITE,
            bg=typer.colors.BLACK,
        )
        raise typer.Exit()


def ensure_solr(version: str = "latest", overwrite: bool = False) -> None:
    """Download and unpack the monarch solr kg, and check permissions."""
    data_path = monarchstow.base / "solr" / "data"
    # When untarred solr won't necessarily use the same file names for index segments etc
    # so the untarred path needs to be removed before updating
    if overwrite:
        shutil.rmtree(data_path, ignore_errors=True)
    monarchstow.ensure_untar(url=f"{KG_DEV_URL}/{version}/solr.tar.gz", force=overwrite)
    if sys.platform in ["linux", "linux2", "darwin"]:
        stat_info = os.stat(data_path)
        if stat_info.st_gid != 8983:
            console.print(
                f"""
Solr container requires write access to {monarchstow.base}.
Please run the following command to set permissions:
    [grey84 on black]sudo chgrp -R 8983 {monarchstow.base} && \ [/]
    [grey84 on black]sudo chmod -R g+w {monarchstow.base}[/]
            """
            )
            sys.exit(1)


def check_for_solr(dc: docker.DockerClient, quiet: bool = False):
    if not quiet:
        console.print("\nChecking for Solr container...")
    c = dc.containers.list(all=True, filters={"name": "monarch_solr"})
    return None if not c else c[0]


def get_solr(update: bool = False):
    """Checks for Solr data and container, and returns a SolrImplementation."""
    if update:
        ensure_solr(update)
    if check_for_solr(dc=docker.from_env(), quiet=True):
        return SolrImplementation()
    else:
        console.print("\nNo Solr container found!\nStart a Solr container with [bold]monarch solr start[/].")
        sys.exit(1)


def start_solr():
    """Starts a local Solr container."""
    data = monarchstow.join("solr", "data")
    dc = docker.from_env()
    c = check_for_solr(dc, quiet=True)
    if not c:
        try:
            c = dc.containers.run(
                "solr:8",
                ports={"8983": 8983},
                volumes=[f"{data}:/opt/solr-data"],
                environment=["SOLR_HOME=/opt/solr-data", "SOLR_OPTS=-Dsolr.jetty.request.header.size=65535"],
                name="monarch_solr",
                command="",
                detach=True,
            )
            time.sleep(10)
            console.print(f"{c.name} {c.status}")
        except Exception as e:
            console.print(f"Error instantiating monarch solr container: {e}")
            raise e
    else:
        try:
            c.start()
        except Exception as e:
            console.print(f"Error running existing container {c.name} ({c.status}) - {e}")
            raise e


def stop_solr():
    """Stops the local Solr container."""
    c = check_for_solr(dc=docker.from_env(), quiet=True)
    if c:
        try:
            console.print(f"Stopping {c.name}...")
            c.stop()
            c.remove()
        except Exception as e:
            console.print(f"Error stopping container {c.name} ({c.status}) - {e}")
            raise e


def solr_status():
    c = check_for_solr(dc=docker.from_env())
    if not c:
        console.print(
            """
No monarch_solr container found. 

Download the Monarch Solr KG and start a local solr instance:
    [grey84 on black]monarch solr start[/]
"""
        )
    else:
        console.print(
            f"""
Found monarch_solr container: {c.id}
Container status: {c.status}
        """
        )
        if c.status == "exited":
            console.print(
                """
Start the container using:
    [grey84 on black]monarch solr start[/]
"""
            )
        if c.status == "running":
            console.print(
                """
You can create a new container with
    [grey84 on black]monarch solr stop[/]
    [grey84 on black]monarch solr start[/]
"""
            )



from abc import ABC

from monarch_py.datamodels.model import Entity


class GroundingInterface(ABC):
    """Abstract interface for grounding text to entities"""

    def ground_entity(self, text: str) -> Entity:
        """
        Grounds a single entity

        Args:
            text (str): Text to ground
        Returns:
            Entity: Dataclass representing a single entity
        """

        raise NotImplementedError



"""Middleware to log requests."""

from loguru import logger
import logging

from fastapi import Request
from fastapi.logger import logger as fastapi_logger
from starlette.middleware.base import BaseHTTPMiddleware

fastapi_logger.setLevel(logging.INFO)


class LoggingMiddleware(BaseHTTPMiddleware):
    """Middleware to log requests."""

    async def dispatch(self, request: Request, call_next):
        """
        Log requests.

        :param request: The request.
        :param call_next: The next call.
        :param logger: The logger.
        :return: The response.
        """
        # Log request method and URL
        logger.info(f"Request URL: {request.url} | Method: {request.method}")

        # Log request headers
        # logger.info(f"Headers: {dict(request.headers)}")

        # If you need the request body, handle with care:
        # body = await request.body()
        # logger.info(f"Body: {body.decode()}")
        #
        # Since the body is read and can't be read again,
        # you need to make it available for the actual route again
        # request._body = body

        response = await call_next(request)
        return response



from typing import List, Optional

from monarch_py.datamodels.solr import HistoPhenoKeys, SolrQuery
from monarch_py.datamodels.category_enums import AssociationPredicate
from monarch_py.utils.association_type_utils import AssociationTypeMappings, get_solr_query_fragment
from monarch_py.utils.utils import escape


def build_association_query(
    category: Optional[List[str]] = None,
    subject: Optional[List[str]] = None,
    subject_closure: Optional[str] = None,
    subject_category: Optional[List[str]] = None,
    subject_namespace: Optional[List[str]] = None,
    subject_taxon: Optional[List[str]] = None,
    predicate: Optional[List[str]] = None,
    object: Optional[List[str]] = None,
    object_closure: Optional[str] = None,
    object_category: Optional[List[str]] = None,
    object_namespace: Optional[List[str]] = None,
    object_taxon: Optional[List[str]] = None,
    entity: Optional[List[str]] = None,
    direct: bool = False,
    q: Optional[str] = None,
    sort: Optional[List[str]] = None,
    facet_fields: Optional[List[str]] = None,
    facet_queries: Optional[List[str]] = None,
    filter_queries: Optional[List[str]] = None,
    facet_mincount: int = 1,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:

    entity_fields = ["subject", "object", "disease_context_qualifier"]
    """Populate a SolrQuery object with association filters"""
    query = SolrQuery(start=offset, rows=limit)
    query.add_field_filter_query("category", None if not category else [c for c in category])
    query.add_field_filter_query("predicate", None if not predicate else [p for p in predicate])
    query.add_field_filter_query("subject_closure", subject_closure)
    query.add_field_filter_query("subject_category", None if not subject_category else [c for c in subject_category])
    query.add_field_filter_query("subject_namespace", subject_namespace)
    query.add_field_filter_query("subject_taxon", subject_taxon)
    query.add_field_filter_query("object_closure", object_closure)
    query.add_field_filter_query("object_category", None if not object_category else [c for c in object_category])
    query.add_field_filter_query("object_namespace", object_namespace)
    query.add_field_filter_query("object_taxon", object_taxon)
    if subject:
        if direct:
            query.add_field_filter_query("subject", " OR ".join(subject))
        else:
            query.add_filter_query(" OR ".join([f'subject:"{s}" OR subject_closure:"{s}"' for s in subject]))
    if object:
        if direct:
            query.add_field_filter_query("object", " OR ".join(object))
        else:
            query.add_filter_query(" OR ".join([f'object:"{o}" OR object_closure:"{o}"' for o in object]))
    if entity:
        if direct:
            query.add_filter_query(" OR ".join([f'{field}:"{e}"' for e in entity for field in entity_fields]))
        else:
            query.add_filter_query(
                " OR ".join([f'{field}:"{e}" OR {field}_closure:"{e}"' for e in entity for field in entity_fields])
            )
    if q:
        # We don't yet have tokenization strategies for the association index, initially we'll limit searching to
        # the visible fields in an association table plus their ID equivalents and use a wildcard query for substring matching
        query.q = q
        query.def_type = "edismax"
        query.hl = True
        query.query_fields = association_search_query_fields()
    if sort:
        query.sort = ", ".join(sort)
    if facet_fields:
        query.facet_fields = facet_fields
    if facet_queries:
        query.facet_queries = facet_queries
    if filter_queries:
        query.filter_queries.extend(filter_queries)
    query.facet_mincount = facet_mincount
    return query


def build_association_table_query(
    entity: List[str],
    category: str,
    direct: bool = False,
    q: Optional[str] = None,
    facet_fields: List[str] = None,
    facet_queries: List[str] = None,
    filter_queries: List[str] = None,
    offset: int = 0,
    limit: int = 5,
    sort: List[str] = None,
) -> SolrQuery:
    if sort is None:
        sort = [
            "frequency_computed_sortable_float desc",
            "evidence_count desc",
            "subject_label asc",
            "predicate asc",
            "object_label asc",
            "primary_knowledge_source asc",
        ]

    query = build_association_query(
        entity=entity,
        category=[category],
        q=q,
        sort=sort,
        offset=offset,
        limit=limit,
        direct=direct,
        facet_fields=facet_fields,
        facet_queries=facet_queries,
        filter_queries=filter_queries,
    )
    return query


def build_association_counts_query(entity: str) -> SolrQuery:
    subject_query = f'AND (subject:"{entity}" OR subject_closure:"{entity}")'
    object_query = f'AND (object:"{entity}" OR object_closure:"{entity}" OR disease_context_qualifier:"{entity}" OR disease_context_qualifier_closure:"{entity}")'

    # Run the same facet_queries constrained to matches against either the subject or object
    # to know which kind of label will be needed in the UI to refer to the opposite side of the association
    facet_queries = []
    for field_query in [subject_query, object_query]:
        for agm in AssociationTypeMappings.get_mappings():
            association_type_query = get_solr_query_fragment(agm)
            facet_queries.append(f"({association_type_query}) {field_query}")
    query = build_association_query(entity=[entity], facet_queries=facet_queries)
    return query


def build_histopheno_query(subject: str) -> SolrQuery:
    query = build_association_query(
        subject=[subject],
        offset=0,
        limit=0,
    )
    hpkeys = [i for i in HistoPhenoKeys]
    query.facet_queries = [f'object_closure:"{(i.value)}"' for i in hpkeys]
    return query


def build_multi_entity_association_query(
    entity: str,
    counterpart_category: Optional[str] = None,
    # predicate: List[str] = None,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:
    """Populate a SolrQuery object with association filters"""
    query = SolrQuery(start=offset, rows=limit)
    if counterpart_category:
        query.add_filter_query(
            f'(subject:"{escape(entity)}" AND object_category:"{escape(counterpart_category)}") OR (object:"{escape(entity)}" AND subject_category:"{escape(counterpart_category)}")'
        )
    else:
        query.add_filter_query(f'(subject:"{escape(entity)}") OR (object:"{escape(entity)}")')
    return query


def build_search_query(
    q: str = "*:*",
    offset: int = 0,
    limit: int = 20,
    category: List[str] = None,
    in_taxon_label: List[str] = None,
    facet_fields: List[str] = None,
    facet_queries: List[str] = None,
    filter_queries: List[str] = None,
    highlighting: bool = False,
    sort: Optional[str] = None,
) -> SolrQuery:
    query = SolrQuery(start=offset, rows=limit, sort=sort)
    query.q = q
    query.def_type = "edismax"
    query.query_fields = entity_query_fields()
    query.hl = highlighting
    query.boost = entity_boost(empty_search=(q == "*:*"))
    if category:
        query.add_filter_query(" OR ".join(f'category:"{cat}"' for cat in category))
    if in_taxon_label:
        query.add_filter_query(" OR ".join([f'in_taxon_label:"{t}"' for t in in_taxon_label]))
    if facet_fields:
        query.facet_fields = facet_fields
    if facet_queries:
        query.facet_queries = facet_queries
    if filter_queries:
        query.filter_queries.extend(filter_queries)
    # Filter out entities that don't have names (required field, but some don't have them)
    query.add_filter_query("name:*")
    return query


def build_autocomplete_query(
    q: str, category: List[str] = None, prioritized_predicates: List[AssociationPredicate] = None
) -> SolrQuery:
    query = SolrQuery(q=q, rows=10, start=0)
    query.q = q
    if category:
        query.add_filter_query(" OR ".join(f'category:"{cat}"' for cat in category))
    # match the query fields to start with
    query.query_fields = entity_query_fields()
    query.def_type = "edismax"
    query.boost = entity_boost(prioritized_predicates=prioritized_predicates, empty_search=(q == "*:*"))
    return query


def build_mapping_query(
    entity_id: Optional[List[str]] = None,
    subject_id: Optional[List[str]] = None,
    predicate_id: Optional[List[str]] = None,
    object_id: Optional[List[str]] = None,
    mapping_justification: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:
    query = SolrQuery(start=offset, rows=limit)
    if entity_id:
        query.add_filter_query(" OR ".join([f'subject_id:"{escape(i)}" OR object_id:"{escape(i)}"' for i in entity_id]))
    if subject_id:
        query.add_filter_query(" OR ".join([f'subject_id:"{escape(i)}"' for i in subject_id]))
    if predicate_id:
        query.add_filter_query(" OR ".join([f'predicate_id:"{escape(i)}"' for i in predicate_id]))
    if object_id:
        query.add_filter_query(" OR ".join([f'object_id:"{escape(i)}"' for i in object_id]))
    if mapping_justification:
        query.add_filter_query(" OR ".join([f'mapping_justification:"{escape(i)}"' for i in mapping_justification]))
    return query


def build_grounding_query(text: str) -> SolrQuery:
    query = SolrQuery(q=text, rows=10, start=0)
    query.q = f'"{text}"'  # quoting so that the complete text is matched as a unit
    # rather than _t (text) or _ac (autocomplete/starts-with), just use keyword fields
    query.query_fields = (
        "id^100 name^10 symbol^10 synonym name_grounding full_name_grounding symbol_grounding synonym_grounding"
    )
    query.def_type = "edismax"
    query.boost = obsolete_unboost(multiplier=0.001)
    return query


### Search helper functions ###


def obsolete_unboost(multiplier=0.1):
    return f'if(termfreq(deprecated,"true"),{multiplier},1)'


def entity_boost(prioritized_predicates: List[AssociationPredicate] = None, empty_search: bool = False) -> str:
    """Shared boost function between search and autocomplete"""
    phenotype_boost = category_boost("biolink:PhenotypicFeature", 1.1)
    disease_boost = category_boost("biolink:Disease", 1.3)
    human_gene_boost = category_boost("biolink:Gene", 1.1, taxon="NCBITaxon:9606")

    boosts = [phenotype_boost, disease_boost, human_gene_boost, obsolete_unboost()]
    if prioritized_predicates:
        boosts.append(entity_predicate_boost(prioritized_predicates, 2.0))
    if empty_search:
        boosts.append(blank_search_boost())
    return f"product({','.join(boosts)})"


def entity_predicate_boost(prioritized_predicates: List[AssociationPredicate], multiplier: float) -> str:
    boosts = []
    for predicate in prioritized_predicates:
        field_root = predicate.value.replace("biolink:", "")
        count_field = field_root + "_count"
        boosts.append(f"if({count_field},{multiplier},1)")
    return ",".join(boosts)


def category_boost(category: str, multiplier: float, taxon: Optional[str] = None) -> str:
    if taxon:
        return f'if(and(termfreq(in_taxon,"{taxon}"),termfreq(category,"{category}")),{multiplier},1)'
    else:
        return f'if(termfreq(category,"{category}"),{multiplier},1)'


def blank_search_boost() -> str:
    """
    Boost specific nodes that we'd like to see as site examples to the top for empty searches
    """
    example_nodes = [
        "MONDO:0007523",  # Ehlers-Danlos syndrome, hypermobility type
        "MONDO:0019391",  # Fanconi anemia
        "MONDO:0018954",  # Loeys-Dietz syndrome
        "MONDO:0011518",  # Wiedemann-Steiner syndrome
        "HP:0001166",  # Arachnodactyly
        "HP:0001631",  # Atrial septal defect
        "UBERON:0000948",  # heart
        "UBERON:0006585",  # vestibular organ
        "HGNC:4851",  # HTT
        "HGNC:3603",  # FBN1
    ]
    # boost score by 1.5 + i/10 for these nodes
    boosts = [f'if(termfreq(id,"{node}"),{len(example_nodes) - i + 2},1)' for i, node in enumerate(example_nodes)]
    boost = ",".join(boosts)
    return f"product({boost})"


def autocomplete_query_fields():
    """
    Fields and boosts used for autocomplete
    """
    return "id^100 name^10 name_t^5 name_ac symbol^10 symbol_t^5 symbol_ac synonym synonym_t synonym_ac"


def entity_query_fields():
    """
    Fields and boosts used for entity search, includes autocomplete fields and expands upon them beyond pure name fields
    """
    return f"{autocomplete_query_fields()} description_t xref"


def association_search_query_fields():
    """
    Shared field list for free text search on associations (e.g. for the association table)
    """

    return (
        "subject subject_label^2 subject_label_t subject_closure subject_closure_label subject_closure_label_t"
        " predicate predicate_t"
        " object object_label^2 object_label_t object_closure object_closure_label object_closure_label_t"
        " publications has_evidence primary_knowledge_source aggregator_knowledge_source provided_by "
    )



from abc import ABC, abstractmethod
from typing import List

from monarch_py.datamodels.model import AssociationResults


class AssociationInterface(ABC):
    """Abstract interface for associations in the Monarch KG"""

    @abstractmethod
    def get_associations(
        self,
        category: List[str] = None,
        subject: List[str] = None,
        predicate: List[str] = None,
        subject_closure: str = None,
        object: List[str] = None,
        object_closure: str = None,
        entity: List[str] = None,
        offset: int = 0,
        limit: int = 20,
    ) -> AssociationResults:
        """Retrieve paginated association records, with filter options

        Args:
            category (str, optional): Filter to only associations matching the specified category. Defaults to None.
            predicate (str, optional): Filter to only associations matching the specified predicate. Defaults to None.
            subject (str, optional): Filter to only associations matching the specified subject. Defaults to None.
            subject_closure (str, optional): Filter to only associations with the specified term ID as an ancestor of the subject. Defaults to None
            object (str, optional): Filter to only associations matching the specified object. Defaults to None.
            object_closure (str, optional): Filter to only associations with the specified term ID as an ancestor of the object. Defaults to None
            entity (str, optional): Filter to only associations where the specified entity is the subject or the object. Defaults to None.
            offset (int, optional): Result offset, for pagination. Defaults to 0.
            limit (int, optional): Limit results to specified number. Defaults to 20.

        Raises:
            NotImplementedError: Use a specific implementation (see the documentation for a list of implementations)

        Returns:
            [AssociationResults](): Dataclass representing results of an association search.
        """
        raise NotImplementedError



import json
from typing import Dict, List

import requests
from loguru import logger
from monarch_py.datamodels.solr import SolrQuery, SolrQueryResult, core
from monarch_py.utils.utils import escape
from pydantic import BaseModel

FIELD_TYPE_SUFFIXES = ["_t", "_ac", "_grounding", "_sortable_float"]


class SolrService(BaseModel):
    base_url: str
    core: core

    def get(self, id):
        url = f"{self.base_url}/{self.core.value}/get?id={id}"
        response = requests.get(url)
        response.raise_for_status()
        entity = response.json()["doc"]
        try:
            self._strip_json(entity, "_version_", "iri")
        except TypeError:  # if entity is None
            return None
        return entity

    def query(self, q: SolrQuery) -> SolrQueryResult:
        url = f"{self.base_url}/{self.core.value}/select?{q.query_string()}"
        response = requests.get(url)
        logger.debug(f"SolrService.query: {url}")
        data = json.loads(response.text)
        if "error" in data:
            logger.error("Solr error message: " + data["error"]["msg"])
        response.raise_for_status()
        solr_query_result = SolrQueryResult.model_validate(data, from_attributes=True)
        solr_query_result.highlighting = SolrService._consolidate_highlights(
            solr_query_result.highlighting, FIELD_TYPE_SUFFIXES
        )
        for doc in solr_query_result.response.docs:
            SolrService._strip_excluded_fields(doc)
            doc["highlighting"] = solr_query_result.highlighting.get(doc["id"], [])

        return solr_query_result

    @staticmethod
    def _strip_excluded_fields(doc: dict):
        SolrService._strip_json(
            doc,
            "_version_",
            "iri",
            "frequency_computed_sortable_float",
            "has_quotient_sortable_float",
            "has_percentage_sortable_float",
        )
        SolrService._strip_json_by_suffix(doc, *FIELD_TYPE_SUFFIXES)

    @staticmethod
    def _strip_json(doc: dict, *fields_to_remove: str):
        for field in fields_to_remove:
            try:
                del doc[field]
            except KeyError:
                pass
        return doc

    @staticmethod
    def _strip_json_by_suffix(doc: dict, *suffixes_to_remove: str):
        for suffix in suffixes_to_remove:
            for key in list(doc.keys()):
                if key.endswith(suffix):
                    del doc[key]

    @staticmethod
    def _consolidate_highlights(
        highlights: Dict[str, Dict[str, List[str]]], suffixes: List[str]
    ) -> Dict[str, Dict[str, List[str]]]:
        """
        For each field that's returned, collapse specified suffix highlighting down to the root field. For example, each
        highlight for name_t and name_ac should be merged into a unique list for name and the specified suffix fields should
        be removed from the highlights.
        """
        consolidated_highlights = highlights.copy()
        for id, highlight in consolidated_highlights.items():
            for field in list(highlight.keys()):
                for suffix in suffixes:
                    if field.endswith(suffix):
                        root_field = field[: -len(suffix)]
                        if root_field not in highlight:
                            highlight[root_field] = highlight[field]
                        else:
                            highlight[root_field] = list(set(highlight[root_field] + highlight[field]))
                            highlight[root_field].sort()
                        del highlight[field]

        return consolidated_highlights

    # Solr returns facet values and counts as a list, they make much more
    # sense as a dictionary
    def _facets_to_dict(self, facet_list: List[str]) -> Dict:
        return dict(zip(facet_list[::2], facet_list[1::2]))

    def get_filtered_facet(self, id, filter_field, facet_field):
        query = SolrQuery(
            rows=0,
            facet=True,
            facet_fields=[facet_field],
            filter_queries=[f"{filter_field}:{escape(id)}"],
        )

        result = self.query(query)

        facet_fields = result.facet_counts.facet_fields[facet_field]

        return self._facets_to_dict(facet_fields)



from enum import Enum


class EntityCategory(Enum):
    """Entity categories"""

    GENE = "biolink:Gene"
    PHENOTYPIC_FEATURE = "biolink:PhenotypicFeature"
    GENOTYPE = "biolink:Genotype"
    ANATOMICAL_ENTITY = "biolink:AnatomicalEntity"
    BIOLOGICAL_PROCESS = "biolink:BiologicalProcess"
    DISEASE = "biolink:Disease"
    NAMED_THING = "biolink:NamedThing"
    PATHWAY = "biolink:Pathway"
    PROTEIN = "biolink:Protein"
    CHEMICAL_ENTITY = "biolink:ChemicalEntity"
    SEQUENCE_VARIANT = "biolink:SequenceVariant"
    MOLECULAR_ACTIVITY = "biolink:MolecularActivity"
    CELLULAR_COMPONENT = "biolink:CellularComponent"
    CELL = "biolink:Cell"
    ORGANISM_TAXON = "biolink:OrganismTaxon"
    MOLECULAR_ENTITY = "biolink:MolecularEntity"
    LIFE_STAGE = "biolink:LifeStage"


class AssociationCategory(Enum):
    """Association categories"""

    PAIRWISE_GENE_TO_GENE_INTERACTION = "biolink:PairwiseGeneToGeneInteraction"
    GENE_TO_EXPRESSION_SITE_ASSOCIATION = "biolink:GeneToExpressionSiteAssociation"
    MACROMOLECULAR_MACHINE_TO_BIOLOGICAL_PROCESS_ASSOCIATION = (
        "biolink:MacromolecularMachineToBiologicalProcessAssociation"
    )
    GENE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:GeneToPhenotypicFeatureAssociation"
    ASSOCIATION = "biolink:Association"
    MACROMOLECULAR_MACHINE_TO_MOLECULAR_ACTIVITY_ASSOCIATION = (
        "biolink:MacromolecularMachineToMolecularActivityAssociation"
    )
    MACROMOLECULAR_MACHINE_TO_CELLULAR_COMPONENT_ASSOCIATION = (
        "biolink:MacromolecularMachineToCellularComponentAssociation"
    )
    GENE_TO_GENE_HOMOLOGY_ASSOCIATION = "biolink:GeneToGeneHomologyAssociation"
    GENOTYPE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:GenotypeToPhenotypicFeatureAssociation"
    DISEASE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:DiseaseToPhenotypicFeatureAssociation"
    GENE_TO_PATHWAY_ASSOCIATION = "biolink:GeneToPathwayAssociation"
    CHEMICAL_TO_PATHWAY_ASSOCIATION = "biolink:ChemicalToPathwayAssociation"
    VARIANT_TO_GENE_ASSOCIATION = "biolink:VariantToGeneAssociation"
    VARIANT_TO_DISEASE_ASSOCIATION = "biolink:VariantToDiseaseAssociation"
    GENOTYPE_TO_DISEASE_ASSOCIATION = "biolink:GenotypeToDiseaseAssociation"
    CORRELATED_GENE_TO_DISEASE_ASSOCIATION = "biolink:CorrelatedGeneToDiseaseAssociation"
    CHEMICAL_TO_DISEASE_OR_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:ChemicalToDiseaseOrPhenotypicFeatureAssociation"
    DISEASE_OR_PHENOTYPIC_FEATURE_TO_GENETIC_INHERITANCE_ASSOCIATION = (
        "biolink:DiseaseOrPhenotypicFeatureToGeneticInheritanceAssociation"
    )
    CAUSAL_GENE_TO_DISEASE_ASSOCIATION = "biolink:CausalGeneToDiseaseAssociation"
    DISEASE_OR_PHENOTYPIC_FEATURE_TO_LOCATION_ASSOCIATION = "biolink:DiseaseOrPhenotypicFeatureToLocationAssociation"
    CHEMICAL_OR_DRUG_OR_TREATMENT_TO_DISEASE_OR_PHENOTYPIC_FEATURE_ASSOCIATION = (
        "biolink:ChemicalOrDrugOrTreatmentToDiseaseOrPhenotypicFeatureAssociation"
    )
    VARIANT_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:VariantToPhenotypicFeatureAssociation"


class AssociationPredicate(Enum):
    """Association predicates"""

    INTERACTS_WITH = "biolink:interacts_with"
    EXPRESSED_IN = "biolink:expressed_in"
    HAS_PHENOTYPE = "biolink:has_phenotype"
    ENABLES = "biolink:enables"
    ACTIVELY_INVOLVED_IN = "biolink:actively_involved_in"
    ORTHOLOGOUS_TO = "biolink:orthologous_to"
    SUBCLASS_OF = "biolink:subclass_of"
    LOCATED_IN = "biolink:located_in"
    RELATED_TO = "biolink:related_to"
    PARTICIPATES_IN = "biolink:participates_in"
    ACTS_UPSTREAM_OF_OR_WITHIN = "biolink:acts_upstream_of_or_within"
    IS_ACTIVE_IN = "biolink:is_active_in"
    PART_OF = "biolink:part_of"
    CAUSES = "biolink:causes"
    IS_SEQUENCE_VARIANT_OF = "biolink:is_sequence_variant_of"
    MODEL_OF = "biolink:model_of"
    ACTS_UPSTREAM_OF = "biolink:acts_upstream_of"
    TREATS_OR_APPLIED_OR_STUDIED_TO_TREAT = "biolink:treats_or_applied_or_studied_to_treat"
    HAS_MODE_OF_INHERITANCE = "biolink:has_mode_of_inheritance"
    GENE_ASSOCIATED_WITH_CONDITION = "biolink:gene_associated_with_condition"
    CONTRIBUTES_TO = "biolink:contributes_to"
    ASSOCIATED_WITH_INCREASED_LIKELIHOOD_OF = "biolink:associated_with_increased_likelihood_of"
    COLOCALIZES_WITH = "biolink:colocalizes_with"
    GENETICALLY_ASSOCIATED_WITH = "biolink:genetically_associated_with"
    DISEASE_HAS_LOCATION = "biolink:disease_has_location"
    ACTS_UPSTREAM_OF_POSITIVE_EFFECT = "biolink:acts_upstream_of_positive_effect"
    ACTS_UPSTREAM_OF_OR_WITHIN_POSITIVE_EFFECT = "biolink:acts_upstream_of_or_within_positive_effect"
    AMELIORATES_CONDITION = "biolink:ameliorates_condition"
    ACTS_UPSTREAM_OF_NEGATIVE_EFFECT = "biolink:acts_upstream_of_negative_effect"
    ACTS_UPSTREAM_OF_OR_WITHIN_NEGATIVE_EFFECT = "biolink:acts_upstream_of_or_within_negative_effect"
    PREVENTATIVE_FOR_CONDITION = "biolink:preventative_for_condition"
    CONTRAINDICATED_IN = "biolink:contraindicated_in"


class MappingPredicate(Enum):
    """Mapping predicates"""

    EXACT_MATCH = "skos:exactMatch"
    CLOSE_MATCH = "skos:closeMatch"
    BROAD_MATCH = "skos:broadMatch"



import oaklib.datamodels.ontology_metadata as omd
from oaklib import OntologyResource
from oaklib.constants import OAKLIB_MODULE
from oaklib.implementations.sqldb.sql_implementation import SqlImplementation

from fastapi import HTTPException

from monarch_py.api.additional_models import SemsimSearchGroup

IS_A = omd.slots.subClassOf.curie
HP_DB_URL = "https://s3.amazonaws.com/bbop-sqlite/hp.db.gz"


def compare_termsets(
    subjects=[""],
    objects=[""],
    predicates=[IS_A, "BFO:0000050"],
    offset: int = 0,
    limit: int = 20,
):
    hp_db = OAKLIB_MODULE.ensure_gunzip(url=HP_DB_URL, autoclean=False)
    oi = SqlImplementation(OntologyResource(slug=hp_db))
    results = oi.termset_pairwise_similarity(subjects, objects, predicates)
    return results


def parse_similarity_prefix(prefix: str):
    if prefix in SemsimSearchGroup._member_names_:
        prefix = prefix
    elif SemsimSearchGroup(prefix):
        prefix = SemsimSearchGroup(prefix).name
    else:
        raise HTTPException(status_code=404, detail="Prefix not found")
    return prefix



from dataclasses import dataclass
from typing import List

import spacy

from monarch_py.interfaces.grounding_interface import GroundingInterface
from monarch_py.interfaces.text_annotation_interface import TextAnnotatorInterface
from monarch_py.datamodels.model import TextAnnotationResult, SearchResult

import pystow
import tarfile


@dataclass
class SpacyImplementation(TextAnnotatorInterface):
    """Implementation of Monarch Interfaces for SPACY"""

    nlp = None
    grounding_implementation = None

    def init_spacy(self, grounding_implementation: GroundingInterface):
        # Define the URL for the Spacy model
        model_url = "https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.5.4/en_core_sci_sm-0.5.4.tar.gz"

        # Use pystow.ensure to download and cache the model archive
        model_archive = pystow.ensure("spacy", "models", url=model_url)
        # Define the expected unpacked directory
        model_dir = model_archive.parent / "en_core_sci"

        # Unpack the model if it's not already unpacked
        if not model_dir.exists():
            print("Unpacking Spacy model...")
            with tarfile.open(model_archive, "r:gz") as tar:
                tar.extractall(path=model_dir.parent)

        model_subdir = next(
            (d for d in model_archive.parent.iterdir() if d.is_dir() and d.name.startswith("en_core_sci")), None
        )

        if model_subdir:
            inner_model_dir = next(
                (
                    d
                    for d in model_subdir.iterdir()
                    if d.is_dir() and d.name.startswith("en_core_sci") and "egg-info" not in d.name
                ),
                None,
            )
            if inner_model_dir:
                # Load the model
                self.nlp = spacy.load(
                    str(str(model_archive.parent / model_subdir.name / inner_model_dir.name / model_subdir.name))
                )

        # Assign the grounding implementation
        self.grounding_implementation = grounding_implementation

        # Test the model with a sample sentence
        self.nlp("Nystagmus, strabismus, fundus, ocular albinism, lewis.")

    def get_annotated_entities(self, text) -> List[TextAnnotationResult]:
        """Annotate text using SPACY"""
        results: List[TextAnnotationResult] = []
        doc = self.nlp(text)

        # Identify tokens to be excluded from annotations
        excluded_tokens = set()
        for token in doc:
            # To filter out tokens representing people's names
            if token.dep_ == "nsubj" and (token.pos_ in {"PROPN", "NOUN"}) and not token.ent_type_:
                excluded_tokens.add(token.text.lower())

        # Modify text to match entity capitalization to handle inconsistent annotations
        for entity in doc.ents:
            matching_entities = self.grounding_implementation.ground_entity(entity.text)
            if matching_entities and text.lower().count(entity.text.lower()) > 1:
                text = text.replace(entity.text, entity.text.title())

        doc = self.nlp(text)  # Recreate the doc after modifying the text

        # Annotate entities
        for entity in doc.ents:
            if entity.text.lower() not in excluded_tokens:
                matching_entities = self.grounding_implementation.ground_entity(entity.text)
                if matching_entities:
                    result = TextAnnotationResult(
                        text=entity.text, tokens=matching_entities, start=entity.start_char, end=entity.end_char
                    )
                    results.append(result)

        # Add non-entity results
        return self.add_non_entity_results(text, results)

    def annotate_text(self, text) -> str:
        """Returns an html formatted string with tags wrapping entities found in the text"""
        annotated_entities = self.get_annotated_entities(text)

        spans = []

        for result in annotated_entities:
            if isinstance(result, TextAnnotationResult):
                if result.tokens:
                    span_text = result.text
                    span_data = ",".join(
                        f"{sr.name},{sr.id},{sr.category}" for sr in result.tokens if isinstance(sr, SearchResult)
                    )
                    spans.append(f'<span class="sciCrunchAnnotation" data-sciGraph="{span_data}">{span_text}</span>')
                else:
                    spans.append(result.text)
            else:
                spans.append(result.text)

        return "".join(spans).rstrip(", ")

    # def ground_entity(self, entity_text: str) -> List[Entity]:
    #     """
    #     Grounds an entity (finds an appropriate ID or a collection of potential identifiers),
    #     using the search engine (Solr), returning the first 3 matches along with any exact matches,
    #     allowing for (hpo) and (mpo) suffixes
    #
    #     """
    #
    #     filtered_search_results = [result for result in search_results.items if not result.deprecated]
    #
    #     returned_entities = filtered_search_results[:3]
    #
    #     for result in filtered_search_results[3:]:
    #         stripped_result_name = result.name.lower().replace(" (hpo)", "").replace(" (mpo)", "")
    #         if entity_text.lower() == stripped_result_name:
    #             returned_entities.append(result)
    #
    #     return returned_entities
    def add_non_entity_results(self, text: str, entities: List[TextAnnotationResult]) -> List[TextAnnotationResult]:
        """Adds non-entity text to the list of entities"""
        results = []
        start_index = 0
        for entity in entities:
            if start_index < entity.start:
                non_span_text = text[start_index : entity.start]
                results.append(TextAnnotationResult(text=non_span_text, start=start_index, end=entity.start))
            results.append(entity)
            start_index = entity.end
        if start_index < len(text):
            non_span_text = text[start_index:]
            results.append(TextAnnotationResult(text=non_span_text, start=start_index, end=len(text)))

        # results.append({"text": "\n"})
        return results



from dataclasses import dataclass
from typing import List, Union

import pystow
from loguru import logger
from pydantic import ValidationError

from monarch_py.datamodels.model import (
    Association,
    CompactAssociation,
    CompactAssociationResults,
    AssociationResults,
    Entity,
    Node,
    NodeHierarchy,
)
from monarch_py.interfaces.association_interface import AssociationInterface
from monarch_py.interfaces.entity_interface import EntityInterface
from monarch_py.service.curie_service import converter
from monarch_py.utils.utils import SQL_DATA_URL, dict_factory

monarchstow = pystow.module("monarch")


@dataclass
class SQLImplementation(EntityInterface, AssociationInterface):
    """Implementation of Monarch Interfaces for SQL endpoint"""

    ###############################
    # Implements: EntityInterface #
    ###############################

    def get_entity(self, id: str, update: bool = False, extra: bool = False) -> Entity:
        """Retrieve a specific entity by exact ID match, writh optional extras

        Args:
            id (str): id of the entity to search for.
            get_association_counts (bool, optional): Whether to get association counts. Defaults to False.
            get_hierarchy (bool, optional): Whether to get the entity hierarchy. Defaults to False.

        Returns:
            Entity: Dataclass representing results of an entity search.
        """

        with monarchstow.ensure_open_sqlite_gz("sql", url=SQL_DATA_URL, force=update) as db:
            db.row_factory = dict_factory
            cur = db.cursor()
            sql_data = cur.execute(f"SELECT * FROM nodes WHERE id = '{id}'").fetchone()

        if not sql_data:
            return None
        results = {
            "id": sql_data["id"],
            "category": sql_data["category"],  # .split("|"), # This will become a list in the future
            "name": sql_data["name"],
            "description": sql_data["description"],
            "xref": sql_data["xref"].split("|"),
            "provided_by": sql_data["provided_by"],
            "in_taxon": sql_data["in_taxon"],
            "symbol": sql_data["symbol"],
            "synonym": sql_data["synonym"].split("|"),
            "uri": converter.expand(sql_data["id"]),
        }
        try:
            results["source"] = sql_data["source"]
        except KeyError:
            pass
        # Convert empty strings to null value
        for k in results:
            results[k] = None if not results[k] else results[k]

        if not extra:
            try:
                return Entity(**results)
            except ValidationError:
                logger.error(f"Validation error for {sql_data}")
                raise

        node = Node(**results)

        if "biolink:Disease" in node.category:
            mode_of_inheritance_associations = self.get_associations(
                subject=id, predicate="biolink:has_mode_of_inheritance", offset=0
            )
        if mode_of_inheritance_associations is not None and len(mode_of_inheritance_associations.items) == 1:
            node.inheritance = self._get_counterpart_entity(mode_of_inheritance_associations.items[0], node)
        node.node_hierarchy = self._get_node_hierarchy(node)
        ### SQL does not support association counts
        return node

    def _get_counterpart_entity(self, association: Association, this_entity: Entity) -> Entity:
        """Returns the other Entity in an Association given this_entity"""
        if this_entity.id in association.subject_closure:
            entity = Entity(
                id=association.object,
                name=association.object_label,
                category=association.object_category if len(association.object_category) == 1 else [],
            )
        elif this_entity.id in association.object_closure:
            entity = Entity(
                id=association.subject,
                name=association.subject_label,
                category=association.subject_category if len(association.subject_category) == 1 else [],
            )
        else:
            raise ValueError(f"Association does not contain this_entity: {this_entity.id}")

        return entity

    def _get_counterpart_entities(
        self,
        this_entity: Entity,
        entity: str = None,
        subject: str = None,
        predicate: str = None,
        object: str = None,
    ) -> List[Entity]:
        """
        Get a list of entities directly associated with this_entity fetched from associations
        in the Solr index

        Args:
            this_entity (Entity): The entity to get associations for
            entity (str, optional): an entity ID occurring in either the subject or predicate. Defaults to None.
            subject (str, optional): an entity ID occurring in the subject. Defaults to None.
            predicate (str, optional): a predicate value. Defaults to None.
            object (str, optional): an entity ID occurring in the object. Defaults to None.
        """
        return [
            self._get_counterpart_entity(association, this_entity)
            for association in self.get_associations(
                entity=entity,
                subject=subject,
                predicate=predicate,
                object=object,
                direct=True,
                offset=0,
            ).items
        ]

    def _get_node_hierarchy(self, entity: Entity) -> NodeHierarchy:
        """
        Get a NodeHierarchy for the given entity

        Args:
            entity (Entity): The entity to get the hierarchy for
            si (SolrInterface): A SolrInterface instance

        Returns:
            NodeHierarchy: A NodeHierarchy object
        """

        super_classes = self._get_counterpart_entities(
            this_entity=entity, subject=entity.id, predicate="biolink:subclass_of"
        )
        sub_classes = self._get_counterpart_entities(
            this_entity=entity, object=entity.id, predicate="biolink:subclass_of"
        )

        return NodeHierarchy(
            super_classes=super_classes,
            sub_classes=sub_classes,
        )

    ####################################
    # Implements: AssociationInterface #
    ####################################

    def get_associations(
        self,
        category: List[str] = None,
        subject: List[str] = None,
        predicate: List[str] = None,
        object: List[str] = None,
        subject_closure: str = None,
        object_closure: str = None,
        entity: List[str] = None,
        direct: bool = None,
        compact: bool = False,
        offset: int = 0,
        limit: int = 20,
    ) -> Union[AssociationResults, CompactAssociationResults]:
        """Retrieve paginated association records, with filter options

        Args:
            category (str, optional): Filter to only associations matching the specified category. Defaults to None.
            predicate (str, optional): Filter to only associations matching the specified predicate. Defaults to None.
            subject (str, optional): Filter to only associations matching the specified subject. Defaults to None.
            object (str, optional): Filter to only associations matching the specified object. Defaults to None.
            subject_closure (str, optional): Filter to only associations with the specified term ID as an ancestor of the subject. Defaults to None.
            object_closure (str, optional): Filter to only associations the specified term ID as an ancestor of the object. Defaults to None.
            entity (str, optional): Filter to only associations where the specified entity is the subject or the object. Defaults to None.
            association_type (str, optional): Filter to only associations matching the specified association label. Defaults to None.
            compact (bool, optional): Whether to return compact or full association records. Defaults to False.
            offset (int, optional): Result offset, for pagination. Defaults to 0.
            limit (int, optional): Limit results to specified number. Defaults to 20.

        Returns:
            AssociationResults: Dataclass representing results of an association search.
        """

        clauses = []
        if category:
            clauses.append(" OR ".join([f"category = '{c}'" for c in category]))
        if subject:
            if direct:
                clauses.append(" OR ".join([f"subject = '{s}'" for s in subject]))
            else:
                clauses.append(" OR ".join([f"subject = '{s}' OR subject_closure like '%{s}%'" for s in subject]))
        if predicate:
            clauses.append(" OR ".join([f"predicate = '{p}'" for p in predicate]))
        if object:
            if direct:
                clauses.append(" OR ".join([f"object = '{o}'" for o in object]))
            else:
                clauses.append(" OR ".join([f"object = '{o}' OR object_closure like '%{o}%'" for o in object]))
        if subject_closure:
            clauses.append(f"subject_closure like '%{subject_closure}%'")
        if object_closure:
            clauses.append(f"object_closure like '%{object_closure}%'")
        if entity:
            if direct:
                clauses.append(" OR ".join([f"subject = '{e}' OR object = '{e}'" for e in entity]))
            else:
                clauses.append(
                    " OR ".join(
                        [
                            f"subject = '{e}' OR object = '{e}' OR subject_closure like '%{e}%' OR object_closure like '%{e}%'"
                            for e in entity
                        ]
                    )
                )

        query = f"SELECT * FROM denormalized_edges "
        if clauses:
            query += "WHERE " + " AND ".join(clauses)
        if limit:
            query += f" LIMIT {limit} OFFSET {offset}"

        count_query = f"SELECT COUNT(*) FROM denormalized_edges "
        if clauses:
            count_query += "WHERE " + " AND ".join(clauses)

        with monarchstow.ensure_open_sqlite_gz("sql", url=SQL_DATA_URL) as db:
            db.row_factory = dict_factory
            cur = db.cursor()
            results = cur.execute(query).fetchall()
            count = cur.execute(count_query).fetchone()
            total = count[f"COUNT(*)"]

        associations = []
        if compact:
            for row in results:
                result = {
                    "category": row["category"],
                    "subject": row["subject"],
                    "subject_label": row["subject_label"],
                    "predicate": row["predicate"],
                    "object": row["object"],
                    "object_label": row["object_label"],
                    "negated": False if not row["negated"] else True,
                }
                # Convert empty strings to null value
                for key in result:
                    result[key] = None if not result[key] else result[key]
                try:
                    associations.append(CompactAssociation(**result))
                except ValidationError:
                    logger.error(f"Validation error for {row}")
                    raise
            return CompactAssociationResults(items=associations, limit=limit, offset=offset, total=total)
        else:
            for row in results:
                result = {
                    "id": row["id"],
                    "original_subject": row["original_subject"],
                    "predicate": row["predicate"],
                    "original_object": row["original_object"],
                    "category": row["category"],
                    "aggregator_knowledge_source": row["aggregator_knowledge_source"].split("|"),
                    "primary_knowledge_source": row["primary_knowledge_source"],
                    "publications": row["publications"].split("|"),
                    "qualifiers": row["qualifiers"].split("|"),
                    "provided_by": row["provided_by"],
                    "has_evidence": row["has_evidence"].split("|"),
                    "stage_qualifier": row["stage_qualifier"],
                    "negated": False if not row["negated"] else True,
                    "frequency_qualifier": row["frequency_qualifier"],
                    "onset_qualifier": row["onset_qualifier"],
                    "sex_qualifier": row["sex_qualifier"],
                    "subject": row["subject"],
                    "object": row["object"],
                }
                # Convert empty strings to null value
                for key in result:
                    result[key] = None if not result[key] else result[key]
                    if isinstance(result[key], list) and len(result[key]) == 1 and not result[key][0]:
                        result[key] = []
                try:
                    associations.append(Association(**result))
                except ValidationError:
                    logger.error(f"Validation error for {row}")
                    raise
            return AssociationResults(items=associations, limit=limit, offset=offset, total=total)



from abc import ABC, abstractmethod
from typing import List

from monarch_py.datamodels.model import AssociationTableResults, FacetValue, SearchResults


class SearchInterface(ABC):
    """Abstract interface for searching the Monarch KG in a Lucene way"""

    @abstractmethod
    def search(
        self,
        q: str,
        offset: int = 0,
        limit: int = 20,
        category: List[str] = None,
        in_taxon: List[str] = None,
        facet_fields: List[str] = None,
        filter_queries: List[str] = None,
        facet_queries: List[str] = None,
        sort: str = None,
    ) -> SearchResults:
        """

        Args:
            q (str): Query string to match against
            category (str): Limit results to only this category
            taxon (str): Limit results to only this taxon
            offset (int): Offset of the first result to return, defaults to 0
            limit (int): Limit the number of results to return, defaults to 20

        Raises:
            NotImplementedError: Use a specific implementation (see the documentation for a list of implementations)

        Returns:
            EntityResults: Dataclass representing results of a generic entity search.
        """
        raise NotImplementedError

    def autocomplete(self, q: str) -> SearchResults:
        """

        Args:
            q (str): Query string to match against

        Raises:
            NotImplementedError: Use a specific implementation (see the documentation for a list of implementations)

        Returns:
            SearchResults: Dataclass representing results of a generic entity search.
        """
        raise NotImplementedError

    def get_association_facets(
        self,
        facet_fields: List[str] = None,
        facet_queries: List[str] = None,
        category: str = None,
        predicate: str = None,
        subject: str = None,
        subject_closure: str = None,
        object: str = None,
        object_closure: str = None,
        entity: str = None,
    ) -> SearchResults:
        """
        Get facet counts and facet query counts for associations
        Args:
            facet_fields (List[str]): Facet fields to return counts for
            facet_queries (List[str]): Facet queries to return counts for
            category (str): Filter to only associations matching the specified category
            predicate (str): Filter to only associations matching the specified predicate
            subject (str): Filter to only associations matching the specified subject
            subject_closure (str): Filter to only associations with the specified term ID as an ancestor of the subject
            object (str): Filter to only associations matching the specified object
            object_closure (str): Filter to only associations with the specified term ID as an ancestor of the object
            entity (str): Filter to only associations where the specified entity is the subject or the object
        Returns:
            SearchResults: Dataclass representing results of a search, with zero rows returned but total count
            and faceting information populated
        """
        raise NotImplementedError

    def get_association_counts(self, entity: str) -> List[FacetValue]:
        """
        Get counts of associations for a given entity
        Args:
            entity (str): Entity to get association counts for
        Returns:
            List[FacetValue]: List of FacetValue objects representing the counts of associations for the given entity
        """
        raise NotImplementedError

    def get_association_table(
        self,
        entity: str,
        category: str,
        query=None,
        sort=None,
        offset=0,
        limit=5,
    ) -> AssociationTableResults:
        """
        Get associations for an entity matching a specified type, with optional search and sort parameters

        Args:
            entity (str): Entity to get associations for
            category (str): Category of associations to return
            query (str): Query string to match against
            sort (str): Sort order, defaults to None
            offset (int): Offset of the first result to return, defaults to 0
            limit (int): Limit the number of results to return, defaults to 20

        Returns:
            AssociationResults: Dataclass representing results of an association search.
        """

        raise NotImplementedError
